// Copyright 2009 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build linux && sw64
// +build linux,sw64

package unix

//TODO: merge from ztypes and zerrors
const (
	BPF_ANY     = 0x0
	BPF_NOEXIST = 0x1
	BPF_EXIST   = 0x2

	BPF_F_NO_PREALLOC    = 0x1
	BPF_F_NO_COMMON_LRU  = 0x2
	BPF_F_NUMA_NODE      = 0x4
	BPF_F_RDONLY         = 0x8
	BPF_F_WRONLY         = 0x10
	BPF_F_STACK_BUILD_ID = 0x20

	BPF_F_RECOMPUTE_CSUM  = 0x1
	BPF_F_INVALIDATE_HASH = 0x2
	BPF_F_HDR_FIELD_MASK  = 0xf
	BPF_F_PSEUDO_HDR      = 0x10
	BPF_F_MARK_MANGLED_0  = 0x20
	BPF_F_MARK_ENFORCE    = 0x40
	BPF_F_INGRESS         = 0x1
	BPF_F_TUNINFO_IPV6    = 0x1
	BPF_F_SKIP_FIELD_MASK = 0xff
	BPF_F_USER_STACK      = 0x100
	BPF_F_FAST_STACK_CMP  = 0x200
	BPF_F_REUSE_STACKID   = 0x400
	BPF_F_USER_BUILD_ID   = 0x800
	BPF_F_ZERO_CSUM_TX    = 0x2
	BPF_F_DONT_FRAGMENT   = 0x4
	BPF_F_SEQ_NUMBER      = 0x8
	BPF_F_INDEX_MASK      = 0xffffffff
	BPF_F_CURRENT_CPU     = 0xffffffff
	BPF_F_CTXLEN_MASK     = 0xfffff00000000

	BPF_SOCK_OPS_RTO_CB_FLAG     = 0x1
	BPF_SOCK_OPS_RETRANS_CB_FLAG = 0x2
	BPF_SOCK_OPS_STATE_CB_FLAG   = 0x4

	BPF_SOCK_OPS_ALL_CB_FLAGS = 0x7

	TCP_BPF_IW            = 0x3e9
	TCP_BPF_SNDCWND_CLAMP = 0x3ea

	BPF_DEVCG_ACC_MKNOD = 0x1
	BPF_DEVCG_ACC_READ  = 0x2
	BPF_DEVCG_ACC_WRITE = 0x4
	BPF_DEVCG_DEV_BLOCK = 0x1
	BPF_DEVCG_DEV_CHAR  = 0x2
)

//generate by ztypes
//TODO:doesn't support for linux 4.19 kernel
//just for compiler passed
type EthtoolDrvinfo struct {
	Cmd          uint32
	Driver       [32]byte
	Version      [32]byte
	Fw_version   [32]byte
	Bus_info     [32]byte
	Erom_version [32]byte
	Reserved2    [12]byte
	N_priv_flags uint32
	N_stats      uint32
	Testinfo_len uint32
	Eedump_len   uint32
	Regdump_len  uint32
}

const SizeofOpenHow = 0x18

const ETHTOOL_GDRVINFO = 0x3

const (
	//generate by handle in zerrors_linux_sw64.go
	//_snyh_TODO: this should be generate by improving build script
	TIOCGWINSZ = 0x40087468
	AF_XDP     = 0x2c //TODO:may create problem
	CAN_J1939  = 0x7  //TODO: may create problem
)

const (
	//ALL OF THIS constants are WORKAROUND, and should be removing
	SYS_UMOUNT2    = SYS_UMOUNT
	SYS_NEWFSTATAT = SYS_FSTATAT64
)

//sysnb getxpid() (pid int, ppid int)
// TODO(snyh):  correct handle Getppid and Getpid
// currently manually remove the implements of Getpid and Getppid
// in zsyscall_linux_sw64.go
func Getpid() (pid int)   { pid, _ = getxpid(); return }
func Getppid() (ppid int) { _, ppid = getxpid(); return }

// TODO(snyh):  correct handle Utime
func Utime(path string, buf *Utimbuf) error {
	tv := [2]Timeval{
		{Sec: buf.Actime},
		{Sec: buf.Modtime},
	}
	return utimes(path, &tv)
}

//sys	Fstat64(fd int, st *Stat_t) (err error)
//sys	Lstat64(path string, st *Stat_t) (err error)
//sys	Stat64(path string, st *Stat_t) (err error)
func Fstat(fd int, st *Stat_t) (err error)      { return Fstat64(fd, st) }
func Lstat(path string, st *Stat_t) (err error) { return Lstat64(path, st) }
func Stat(path string, st *Stat_t) (err error)  { return Stat64(path, st) }

//sys getxuid() (uid int, euid int)
func Getuid() (uid int)   { uid, _ = getxuid(); return }
func Geteuid() (euid int) { _, euid = getxuid(); return }

//sys getxgid() (gid int, egid int)
func Getgid() (gid int)   { gid, _ = getxgid(); return }
func Getegid() (egid int) { _, egid = getxgid(); return }

//sys	Statfs(path string, buf *Statfs_t) (err error)
//sys	Fstatfs(fd int, buf *Statfs_t) (err error)
//sys	dup2(oldfd int, newfd int) (err error)
//sys	Fchown(fd int, uid int, gid int) (err error)
//sys	Ftruncate(fd int, length int64) (err error)
//sysnb	Getrlimit(resource int, rlim *Rlimit) (err error)
//sys	Lchown(path string, uid int, gid int) (err error)
//sys	Listen(s int, n int) (err error)
//sys	Pread(fd int, p []byte, offset int64) (n int, err error) = SYS_PREAD64
//sys	Pwrite(fd int, p []byte, offset int64) (n int, err error) = SYS_PWRITE64
//sys	Seek(fd int, offset int64, whence int) (off int64, err error) = SYS_LSEEK
//sys	sendfile(outfd int, infd int, offset *int64, count int) (written int, err error)
//sys	setfsgid(gid int) (prev int, err error)
//sys	setfsuid(uid int) (prev int, err error)
//sysnb	Setregid(rgid int, egid int) (err error)
//sysnb	Setresgid(rgid int, egid int, sgid int) (err error)
//sysnb	Setresuid(ruid int, euid int, suid int) (err error)
//sysnb	Setrlimit(resource int, rlim *Rlimit) (err error)
//sysnb	Setreuid(ruid int, euid int) (err error)
//sys	Shutdown(fd int, how int) (err error)
//sys	Splice(rfd int, roff *int64, wfd int, woff *int64, len int, flags int) (n int64, err error)

//sys	SyncFileRange(fd int, off int64, n int64, flags int) (err error)
//sys	Truncate(path string, length int64) (err error)
//sys	accept(s int, rsa *RawSockaddrAny, addrlen *_Socklen) (fd int, err error)
//sys	accept4(s int, rsa *RawSockaddrAny, addrlen *_Socklen, flags int) (fd int, err error)
//sys	bind(s int, addr unsafe.Pointer, addrlen _Socklen) (err error)
//sys	connect(s int, addr unsafe.Pointer, addrlen _Socklen) (err error)
//sysnb	getgroups(n int, list *_Gid_t) (nn int, err error)
//sysnb	setgroups(n int, list *_Gid_t) (err error)
//sys	getsockopt(s int, level int, name int, val unsafe.Pointer, vallen *_Socklen) (err error)
//sys	setsockopt(s int, level int, name int, val unsafe.Pointer, vallen uintptr) (err error)
//sysnb	socket(domain int, typ int, proto int) (fd int, err error)
//sysnb	socketpair(domain int, typ int, proto int, fd *[2]int32) (err error)
//sysnb	getpeername(fd int, rsa *RawSockaddrAny, addrlen *_Socklen) (err error)
//sysnb	getsockname(fd int, rsa *RawSockaddrAny, addrlen *_Socklen) (err error)
//sys	recvfrom(fd int, p []byte, flags int, from *RawSockaddrAny, fromlen *_Socklen) (n int, err error)
//sys	sendto(s int, buf []byte, flags int, to unsafe.Pointer, addrlen _Socklen) (err error)
//sys	recvmsg(s int, msg *Msghdr, flags int) (n int, err error)
//sys	sendmsg(s int, msg *Msghdr, flags int) (n int, err error)
//sys	mmap(addr uintptr, length uintptr, prot int, flags int, fd int, offset int64) (xaddr uintptr, err error)

//sysnb	Gettimeofday(tv *Timeval) (err error)

func Time(t *Time_t) (tt Time_t, err error) {
	var tv Timeval
	err = Gettimeofday(&tv)
	if err != nil {
		return 0, err
	}
	if t != nil {
		*t = Time_t(tv.Sec)
	}
	return Time_t(tv.Sec), nil
}

func setTimespec(sec, nsec int64) Timespec {
	return Timespec{Sec: sec, Nsec: nsec}
}

func setTimeval(sec, usec int64) Timeval {
	return Timeval{Sec: sec, Usec: usec}
}

func Pipe(p []int) (err error) {
	if len(p) != 2 {
		return EINVAL
	}
	var pp [2]_C_int
	err = pipe2(&pp, 0)
	p[0] = int(pp[0])
	p[1] = int(pp[1])
	return
}

//sysnb pipe2(p *[2]_C_int, flags int) (err error)

func Pipe2(p []int, flags int) (err error) {
	if len(p) != 2 {
		return EINVAL
	}
	var pp [2]_C_int
	err = pipe2(&pp, flags)
	p[0] = int(pp[0])
	p[1] = int(pp[1])
	return
}

func Ioperm(from int, num int, on int) (err error) {
	return ENOSYS
}

func Iopl(level int) (err error) {
	return ENOSYS
}

// func (r *PtraceRegs) PC() uint64 { return r.Epc }
// func (r *PtraceRegs) SetPC(pc uint64) { r.Epc = pc }

func (iov *Iovec) SetLen(length int) {
	iov.Len = uint64(length)
}

func (msghdr *Msghdr) SetControllen(length int) {
	msghdr.Controllen = uint64(length)
}

func (cmsg *Cmsghdr) SetLen(length int) {
	cmsg.Len = uint64(length)
}

//sys poll(fds *PollFd, nfds int, timeout int) (n int, err error)
func Poll(fds []PollFd, timeout int) (n int, err error) {
	if len(fds) == 0 {
		return poll(nil, 0, timeout)
	}
	return poll(&fds[0], len(fds), timeout)
}

//sysnb  EpollCreate(size int) (fd int, err error)
//sys EpollWait(epfd int, events []EpollEvent, msec int) (n int, err error)
//sys Fadvise(fd int, offset int64, length int64, advice int) (err error) = SYS_FADVISE64
//sys Fstatat(dirfd int, path string, stat *Stat_t, flags int) (err error) = SYS_NEWFSTATAT
func Select(nfd int, r *FdSet, w *FdSet, e *FdSet, timeout *Timeval) (n int, err error) {
	var ts *Timespec
	if timeout != nil {
		ts = &Timespec{Sec: timeout.Sec, Nsec: timeout.Usec * 1000}
	}
	return Pselect(nfd, r, w, e, ts, nil)
}

//sys Ustat(dev int, ubuf *Ustat_t) (err error)
//sys futimesat(dirfd int, path string, times *[2]Timeval) (err error)
//sys utimes(path string, times *[2]Timeval) (err error)

//sys	Renameat(olddirfd int, oldpath string, newdirfd int, newpath string) (err error)
